%%% @doc
%%% Clutch Controller
%%%
%%% This process is a in charge of maintaining the program's state.
%%% @end

-module(gmc_con).
-vsn("0.1.0").
-author("Craig Everett <craigeverett@qpq.swiss>").
-copyright("QPQ AG <info@qpq.swiss>").
-license("GPL-3.0-or-later").

-behavior(gen_server).
-export([open_wallet/2, close_wallet/0, password/2,
         make_key/6, recover_key/1, mnemonic/1, rename_key/2, drop_key/1]).
-export([encrypt/2, decrypt/2]).
-export([start_link/0, stop/0, save/1]).
-export([init/1, terminate/2, code_change/3,
         handle_call/3, handle_cast/2, handle_info/2]).
-include("$zx_include/zx_logger.hrl").

-include("gmc.hrl").


%%% Type and Record Definitions


-record(s,
        {version = 1         :: integer(),
         window  = none      :: none | wx:wx_object(),
         wallet  = #wallet{} :: #wallet{},
         prefs   = #{}       :: #{atom() := term()}}).


-type state()   :: #s{}.



%% Interface


-spec open_wallet(Path, Password) -> ok
    when Path     :: file:filename(),
         Password :: string().

open_wallet(Path, Password) ->
    gen_server:cast(?MODULE, {open_wallet, Path, Password}).


-spec close_wallet() -> ok.

close_wallet() ->
    gen_server:cast(?MODULE, close_wallet).


-spec password(Old, New) -> ok
    when Old :: none | string(),
         New :: none | string().

password(Old, New) ->
    gen_server:cast(?MODULE, {password, Old, New}).


-spec make_key(Type, Size, Name, Seed, Encoding, Transform) -> ok
    when Type      :: {eddsa, ed25519},
         Size      :: 256,
         Name      :: string(),
         Seed      :: string(),
         Encoding  :: utf8 | base64 | base58,
         Transform :: raw | {Algo, Yugeness},
         Algo      :: sha3 | sha2 | x_or | pbkdf2,
         Yugeness  :: rand | non_neg_integer().
%% @doc
%% Generate a new key.
%% The magic first two args are here because `ak_*' basically has no option other than
%% to mean a 256 bit key based on Curve 25519. When more key varieties are added to the
%% system this will change quite a lot.

make_key({eddsa, ed25519}, 256, Name, Seed, Encoding, Transform) ->
    gen_server:cast(?MODULE, {make_key, Name, Seed, Encoding, Transform}).


-spec recover_key(Mnemonic) -> ok
    when Mnemonic :: string().

recover_key(Mnemonic) ->
    gen_server:cast(?MODULE, {recover_key, Mnemonic}).


-spec mnemonic(ID) -> {ok, Mnemonic} | error
    when ID       :: clutch:id(),
         Mnemonic :: string().

mnemonic(ID) ->
    gen_server:call(?MODULE, {mnemonic, ID}).


-spec rename_key(ID, NewName) -> ok
    when ID      :: clutch:id(),
         NewName :: string().

rename_key(ID, NewName) ->
    gen_server:cast(?MODULE, {rename_key, ID, NewName}).


-spec drop_key(ID) -> ok
    when ID :: clutch:id().

drop_key(ID) ->
    gen_server:cast(?MODULE, {drop_key, ID}).


-spec stop() -> ok.

stop() ->
    gen_server:cast(?MODULE, stop).


-spec save(Prefs) -> ok | {error, Reason}
    when Prefs  :: #{atom() := term()},
         Reason :: file:posix().

save(Prefs) ->
    gen_server:call(?MODULE, {save, Prefs}).



%%% Startup Functions


-spec start_link() -> Result
    when Result :: {ok, pid()}
                 | {error, Reason},
         Reason :: {already_started, pid()}
                 | {shutdown, term()}
                 | term().
%% @private
%% Called by gmc_sup.

start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, none, []).


-spec init(none) -> {ok, state()}.

init(none) ->
    ok = log(info, "Starting"),
    Prefs = read_prefs(),
    Window = gmc_gui:start_link(Prefs),
    ok = gmc_gui:ask_password(),
    State = #s{window = Window, prefs = Prefs},
    {ok, State}.


read_prefs() ->
    case file:consult(prefs_path()) of
        {ok, Prefs} ->
            proplists:to_map(Prefs);
        _ ->
            #{selected => 0,
              last     => save_path(),
              lang     => en_us,
              geometry => none}
    end.



%%% gen_server Message Handling Callbacks


-spec handle_call(Message, From, State) -> Result
    when Message  :: term(),
         From     :: {pid(), reference()},
         State    :: state(),
         Result   :: {reply, Response, NewState}
                   | {noreply, State},
         Response :: ok
                   | {error, {listening, inet:port_number()}},
         NewState :: state().
%% @private
%% The gen_server:handle_call/3 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_call-3

handle_call({save, Prefs}, _, State) ->
    Response = do_save(State#s{prefs = Prefs}),
    {reply, Response, State};
handle_call({mnemonic, ID}, _, State) ->
    Response = do_mnemonic(ID, State),
    {reply, Response, State};
handle_call(Unexpected, From, State) ->
    ok = log(warning, "Unexpected call from ~tp: ~tp~n", [From, Unexpected]),
    {noreply, State}.


-spec handle_cast(Message, State) -> {noreply, NewState}
    when Message  :: term(),
         State    :: state(),
         NewState :: state().
%% @private
%% The gen_server:handle_cast/2 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_cast-2

handle_cast({make_key, Name, Seed, Encoding, Transform}, State) ->
    NewState = do_make_key(Name, Seed, Encoding, Transform, State),
    {noreply, NewState};
handle_cast({recover_key, Mnemonic}, State) ->
    NewState = do_recover_key(Mnemonic, State),
    {noreply, NewState};
handle_cast({rename_key, ID, NewName}, State) ->
    NewState = do_rename_key(ID, NewName, State),
    {noreply, NewState};
handle_cast({drop_key, ID}, State) ->
    NewState = do_drop_key(ID, State),
    {noreply, NewState};
handle_cast({open_wallet, Path, Phrase}, State) ->
    NewState = do_open_wallet(Path, Phrase, State),
    {noreply, NewState};
handle_cast({password, Old, New}, State) ->
    NewState = do_password(Old, New, State),
    {noreply, NewState};
handle_cast(stop, State) ->
    ok = zx:stop(),
    {noreply, State};
handle_cast(Unexpected, State) ->
    ok = tell(warning, "Unexpected cast: ~tp~n", [Unexpected]),
    {noreply, State}.


-spec handle_info(Message, State) -> {noreply, NewState}
    when Message  :: term(),
         State    :: state(),
         NewState :: state().
%% @private
%% The gen_server:handle_info/2 callback.
%% See: http://erlang.org/doc/man/gen_server.html#Module:handle_info-2

handle_info(Unexpected, State) ->
    ok = log(warning, "Unexpected info: ~tp~n", [Unexpected]),
    {noreply, State}.



%% @private
%% gen_server callback to handle state transformations necessary for hot
%% code updates. This template performs no transformation.

code_change(_, State, _) ->
    {ok, State}.


terminate(normal, _) ->
    zx:stop();
terminate(Reason, State) ->
    ok = log(info, "Reason: ~tp, State: ~tp", [Reason, State]),
    zx:stop().



%%%

do_make_key(Name, <<>>, _, Transform, State) ->
    Bin = crypto:strong_rand_bytes(32),
    do_make_key2(Name, Bin, Transform, State);
do_make_key(Name, Seed, utf8, Transform, State) ->
    Bin = unicode:characters_to_binary(Seed),
    do_make_key2(Name, Bin, Transform, State);
do_make_key(Name, Seed, base64, Transform, State) ->
    case base64_decode(Seed) of
        {ok, Bin} ->
            do_make_key2(Name, Bin, Transform, State);
        {error, Reason} ->
            ok = gmc_gui:trouble({error, {base64, Reason}}),
            State
    end;
do_make_key(Name, Seed, base58, Transform, State) ->
    case base58:check_base58(Seed) of
        true  ->
            Bin = base58:base58_to_binary(Seed),
            do_make_key2(Name, Bin, Transform, State);
        false ->
            ok = gmc_gui:trouble({error, {base58, badarg}}),
            State
    end.

do_make_key2(Name, Bin, Transform, State = #s{wallet = W}) ->
    #wallet{poas = POAs, keys = Keys} = W,
    T = transform(Transform),
    Seed = T(Bin),
    Key = #key{name = KeyName, id = ID} = gmc_key_master:make_key(Name, Seed),
    POA = #poa{name = KeyName, id = ID},
    NewKeys = [Key | Keys],
    NewPOAs = [POA | POAs],
    NewWallet = W#wallet{poas = NewPOAs, keys = NewKeys},
    ok = gmc_gui:show(NewPOAs),
    State#s{wallet = NewWallet}.


base64_decode(String) ->
    try
        {ok, base64:decode(String)}
    catch
        E:R -> {E, R}
    end.


transform({sha3, 256}) ->
    fun(D) -> crypto:hash(sha3_256, D) end;
transform({sha2, 256}) ->
    fun(D) -> crypto:hash(sha256, D) end;
transform({x_or, 256}) ->
    fun t_xor/1;
transform(raw) ->
    fun(D) -> D end.


t_xor(Bin) -> t_xor(Bin, <<0:256>>).

t_xor(<<H:32/binary, T/binary>>, A) ->
    t_xor(T, crypto:exor(H, A));
t_xor(<<>>, A) ->
    A;
t_xor(B, A) ->
    H = <<0:(256 - bit_size(B)), B/binary>>,
    crypto:exor(H, A).


do_recover_key(Mnemonic, State) ->
    case gmc_key_master:decode(Mnemonic) of
        {ok, Seed} ->
            do_recover_key2(Seed, State);
        Error ->
            ok = gmc_gui:trouble(Error),
            State
    end.

do_recover_key2(Seed, State = #s{wallet = W}) ->
    #wallet{keys = Keys, poas = POAs} = W,
    Recovered = #key{id = ID, name = Name} = gmc_key_master:make_key("", Seed),
    case lists:keymember(ID, #key.id, Keys) of
        false ->
            NewKeys = [Recovered | Keys],
            POA = #poa{name = Name, id = ID},
            NewPOAs = [POA | POAs],
            ok = gmc_gui:show(NewPOAs),
            NewWallet = W#wallet{poas = NewPOAs, keys = NewKeys},
            State#s{wallet = NewWallet};
        true ->
            State
    end.


do_mnemonic(ID, #s{wallet = #wallet{keys = Keys}}) ->
    case lists:keyfind(ID, #key.id, Keys) of
        #key{pair = #{secret := <<K:32/binary, _/binary>>}} ->
            Mnemonic = gmc_key_master:encode(K),
            {ok, Mnemonic};
        false ->
            {error, bad_key}
    end.


do_rename_key(ID, NewName, State = #s{wallet = W}) ->
    #wallet{poas = POAs, keys = Keys} = W,
    A = lists:keyfind(ID, #poa.id, POAs),
    K = lists:keyfind(ID, #key.id, Keys),
    NewPOAs = lists:keystore(ID, #poa.id, POAs, A#poa{name = NewName}),
    NewKeys = lists:keystore(ID, #key.id, Keys, K#key{name = NewName}),
    NewWallet = W#wallet{poas = NewPOAs, keys = NewKeys},
    ok = gmc_gui:show(NewPOAs),
    State#s{wallet = NewWallet}.


do_drop_key(ID, State = #s{wallet = W}) ->
    #wallet{poas = POAs, keys = Keys} = W,
    NewPOAs = lists:keydelete(ID, #poa.id, POAs),
    NewKeys = lists:keydelete(ID, #key.id, Keys),
    NewWallet = W#wallet{poas = NewPOAs, keys = NewKeys},
    ok = gmc_gui:show(NewPOAs),
    State#s{wallet = NewWallet}.


encrypt(Pass, Binary) ->
    Flags = [{encrypt, true}, {padding, pkcs_padding}],
    crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags).


decrypt(Pass, Binary) ->
    Flags = [{encrypt, false}, {padding, pkcs_padding}],
    crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags).


pass(Phrase) ->
    crypto:hash(sha3_256, Phrase).


do_open_wallet(Path, none, State) ->
    case read(Path, none) of
        {ok, Recovered = #wallet{poas = POAs, chain_id = ChainID, endpoint = Node}} ->
            ok = gmc_gui:show(POAs),
            ok = gmc_gui:chain(ChainID, Node),
            State#s{wallet = Recovered};
        Error ->
            ok = gmc_gui:trouble(Error),
            State
    end; 
do_open_wallet(Path, Phrase, State) ->
    Pass = pass(Phrase),
    case read(Path, Pass) of
        {ok, Recovered = #wallet{poas = POAs, chain_id = ChainID, endpoint = Node}} ->
            ok = gmc_gui:show(POAs),
            ok = gmc_gui:chain(ChainID, Node),
            State#s{wallet = Recovered};
        Error ->
            ok = gmc_gui:trouble(Error),
            State
    end.


do_password(none, none, State) ->
    State;
do_password(none, New, State = #s{wallet = #wallet{pass = none}}) ->
    Pass = pass(New),
    State#s{wallet = #wallet{pass = Pass}};
do_password(Old, none, State = #s{wallet = #wallet{pass = Pass}}) ->
    case pass(Old) =:= Pass of
        true  -> State#s{wallet = #wallet{pass = none}};
        false -> State
    end;
do_password(Old, New, State = #s{wallet = #wallet{pass = Pass}}) ->
    case pass(Old) =:= Pass of
        true  -> State#s{wallet = #wallet{pass = pass(New)}};
        false -> State
    end.


do_spend(KeyID, TX, State = #s{wallet = #wallet{keys = Keys}}) ->
    case lists:keyfind(KeyID, #key.id, Keys) of
        #key{pair = #{secret := PrivKey}} ->
            do_spend2(PrivKey, TX, State);
        false ->
            log(warning, "Tried do_spend with a bad key: ~p", [KeyID])
    end.

do_spend2(PrivKey,
          #spend_tx{sender_id    = SenderID,
                    recipient_id = RecipientId,
                    amount       = Amount,
                    gas_price    = GasPrice,
                    gas          = Gas,
                    ttl          = TTL,
                    nonce        = Nonce,
                    payload      = Payload},
          #s{wallet = #wallet{chain_id = ChainID}}) ->
    Type = spend_tx,
    Vsn = 1,
    Fields =
        [{sender_id,    SenderID},
         {recipient_id, RecipientId},
         {amount,       Amount},
         {gas_price,    GasPrice},
         {gas,          Gas},
         {ttl,          TTL},
         {nonce,        Nonce},
         {payload,      Payload}],
    Template =
        [{sender_id,    id},
         {recipient_id, id},
         {amount,       int},
         {gas_price,    int},
         {gas,          int},
         {ttl,          int},
         {nonce,        int},
         {payload, binary}],
    BinaryTX  = aeser_chain_objects:serialize(Type, Vsn, Template, Fields),
    NetworkTX = <<ChainID/binary, BinaryTX/binary>>,
    Signature = ecu_eddsa:sign_detached(NetworkTX, PrivKey),
    SigTxType = signed_tx,
    SigTxVsn  = 1,
    SigTemplate =
        [{signatures, [binary]},
         {transaction, binary}],
    TX_Data =
        [{signatures, [Signature]},
         {transaction, NetworkTX}],
    SignedTX = aeser_chain_objects:serialize(SigTxType, SigTxVsn, SigTemplate, TX_Data),
    tell("SpendTX: ~p", [SignedTX]).


do_save(State = #s{prefs = Prefs}) ->
    ok = persist(Prefs),
    do_save2(State).


do_save2(#s{wallet = W = #wallet{pass = none}}) ->
    Path = save_path(),
    ok = filelib:ensure_dir(Path),
    file:write_file(Path, term_to_binary(W));
do_save2(#s{wallet = W = #wallet{pass = Pass}}) ->
    Path = save_path(),
    ok = filelib:ensure_dir(Path),
    Cipher = encrypt(Pass, term_to_binary(W)),
    file:write_file(Path, Cipher).


read(Path, none) ->
    case file:read_file(Path) of
        {ok, Bin} -> read2(Bin);
        Error     -> Error
    end;
read(Path, Pass) ->
    case file:read_file(Path) of
        {ok, Cipher} ->
            try
                Bin = decrypt(Pass, Cipher),
                read2(Bin)
            catch
                error:{error, L, "Can't finalize"} ->
                    ok = log(info, "Decrypt failed at ~p", [L]),
                    {error, bad_password};
                E:R ->
                    tell("Here: ~p", [{E, R}]),
                    {E, R}
            end;
        Error ->
            Error
    end.

read2(Bin) ->
    case zx_lib:b_to_t(Bin) of
        {ok, T} -> {ok, T};
        error   -> {error, badarg}
    end.


save_path() ->
    filename:join(zx_lib:path(var, "otpr", "clutch"), "default.gaju").


persist(Prefs) ->
    Path = prefs_path(),
    ok = filelib:ensure_dir(Path),
    zx_lib:write_terms(Path, proplists:from_map(Prefs)).


prefs_path() ->
    filename:join(zx_lib:path(etc, "otpr", "clutch"), "prefs.eterms").
